function [IC, FC] = detect_GE_Zeni_Enh_Ghoussayni(Data, headerTraj, FootMrk, side, fs, mrkDyn, fig, AnatomicalAxesNotDef, MrkGaps)

%% 1st approximation: GE detection based on Zeni et al. (2008)
% Zeni Jr, J. A., Richards, J. G., & Higginson, J. S. (2008).
% Two simple methods for determining gait events during treadmill and
% overground walking using kinematic data.
% Gait & posture, 27(4), 710-714.
% https://doi.org/10.1016/j.gaitpost.2007.07.007
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

%% Refinement: GE detection based on 3D foot velocities
% Enhanced algorithm based on Ghoussayni et al. (2004) and Bruening & Ridge (2014)

% Ghoussayni, S., Stevens, C., Durham, S., & Ewins, D. (2004).
% Assessment and validation of a simple automated method for
% the detection of gait events and intervals.
% Gait & Posture, 20(3), 266-272.
% https://doi.org/10.1016/j.gaitpost.2003.10.001

%% Modified as suggested in
% Bruening, D. A., & Ridge, S. T. (2014).
% Automated event detection algorithms in pathological gait.
% Gait & posture, 39(1), 472-477.
% https://doi.org/10.1016/j.gaitpost.2013.08.023
%~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

%% 1st approximation: GE detection based on Zeni et al. (2008)
[IC, FC, TrajH, TrajT] = GE_Zeni(Data, headerTraj,FootMrk, side, fs, mrkDyn, AnatomicalAxesNotDef, MrkGaps);

%% Refinement: GE detection based on 3D foot velocities
% Estimate of walking speed (WS)
twindow = 25;
[GEsWS.HS.right, GEsWS.TO.right] = GE_Ghoussayni_fixThr(Data, headerTraj, FootMrk, 'R', fs, twindow, false);
[GEsWS.HS.left, GEsWS.TO.left] = GE_Ghoussayni_fixThr(Data, headerTraj, FootMrk, 'L', fs, twindow, false);

mrk_to_use = 'HEEL';
strideCostr.max_st = 3;
strideCostr.min_st = 0.2;
strideCostr.min_sl = 0.15;
strideCostr.max_h = 0.15;
RWS = strideDet(GEsWS, strideCostr, Data, headerTraj, fs, mrk_to_use);

% GE detection
if ~isempty(RWS.R) && ~isempty(RWS.L)
    [GE_gh.HS, GE_gh.TO] = GE_Ghoussayni_variablethr_RWS(Data, headerTraj, FootMrk, side, fs, RWS, twindow, false);
    
    N = size(Data.(headerTraj).(strcat(side,FootMrk{1})),1);
    HEEL = Data.(headerTraj).(strcat(side,FootMrk{1}));
    TOE = Data.(headerTraj).(strcat(side,FootMrk{2}));
    for t = 1:N-1
        % Heel markers velocity
        Hvelocity(t,1) = sqrt((HEEL(t+1,1)- HEEL(t,1))^2+ ...
            (HEEL(t+1,2)- HEEL(t,2))^2 +...
            (HEEL(t+1,3)- HEEL(t,3))^2)/ ...
            (1/fs);
        
        % Toe markers velocity
        Tvelocity(t,1) = sqrt((TOE(t+1,1)- TOE(t,1)).^2+ ...
            (TOE(t+1,2)- TOE(t,2)).^2 +...
            (TOE(t+1,3)- TOE(t,3)).^2)/ ...
            (1/fs);
    end
    
    % Refine the detected ICs
    IC_G = GE_gh.HS;
    if ~isempty(IC)
        for i = 1:length(IC)
            % match the relevant FC
            [val, posToMatch] = min(abs(IC(i)-IC_G(:,1)));
            if IC_G(posToMatch,1)>IC(i) && val<twindow
                IC(i) = IC_G(posToMatch,1);
            else
                
            end
        end
    else
        
    end
    
    %% Refine with the heel mrk velocity
    FC_G = GE_gh.TO;
    if ~isempty(FC_G)
        FC_G(:,end+1)=zeros(size(FC_G));
        for i = 1:size(FC_G,1)
            if FC_G(i,1)-(twindow/5) > 0 && FC_G(i,1)+(twindow/5) <length(Hvelocity)
                sigNow = Hvelocity(FC_G(i,1)-(twindow/5):FC_G(i,1)+(twindow/5));
                [~,posNow] = findpeaks(sigNow);
                if ~isempty(posNow)
                    if length(posNow)==1
                        FC_G(i,1) = FC_G(i,1)+posNow-(twindow/5)+1;
                        FC_G(i,2)=1;
                    end
                else
                    
                end
            end
        end
    else
        
    end
    
    % Refine FC values
    if ~isempty(FC)
        FC(:,2) = nan(size(FC,1),1);
        for i = 1:length(FC)
            % match the relevant FC
            [~, posToMatch] = min(abs(FC(i,1)-FC_G(:,1)));
            if abs(FC(i,1)-FC_G(posToMatch,1))<twindow
                if FC_G(posToMatch,2)==1 && FC_G(posToMatch,1)>FC(i,1)
                    %                 FC(i,1) = FC_G(posToMatch,1);
                    FC(i,2) = FC_G(posToMatch,1);
                else
                    
                end
            end
        end
    else
    end
    if size(unique(FC(~isnan(FC(:,2)),2)),1) == size(FC(~isnan(FC(:,2)),2),1)
        % OK
        for k = 1:size(FC,1)
            if ~isnan(FC(k,2))
                FC(k,1) = FC(k,2);
            end
        end
        FC = FC(:,1);
    else
        % Check if there are repetitions
    end
    
    if isempty(IC)
        if ~isempty(IC_G)
            IC = IC_G;
        end
    end
    if isempty(FC)
        if ~isempty(FC_G)
            FC = FC_G(:,1);
        end
    end
    
    if ~isempty(IC) && ~isempty(FC)
        %% check Local mrk traj velocity
        vTrajH = diff(TrajH)*fs;
        vTrajT = diff(TrajT)*fs;
        
        %% check portion of the signal with spikes in foot velocity --> possible pelvis mrk swap!
        if range(vTrajH)>5 % there are spikes due to mrk swap!
            vIC = nan(size(IC));
            for k = 1:size(IC,1)
                vIC(k) = vTrajH(IC(k));
            end
            vFC = nan(size(FC));
            for k = 1:size(FC,1)
                vFC(k) = vTrajT(FC(k));
            end
            
            x1 = find(vTrajH<-2);
            u =  find(vTrajT<-2);
            x1 = [x1;u]; x1 = unique(x1);
            PortionToCheck = [];
            if ~isempty(x1)
                s = find(diff(x1)>1);
                if isempty(s)
                    PortionToCheck = [PortionToCheck; x1(1),x1(end)];
                else
                    for i = 1:size(s,1)
                        if i == 1
                            PortionToCheck = [PortionToCheck; x1(1),x1(s(i))];
                        else
                            PortionToCheck = [PortionToCheck; x1(s(i-1)+1),x1(s(i))];
                        end
                    end
                end
            end
            
            % check if this portions are in the parts of the signal where
            % there are gaps --> these would be removed!!!
            if ~isempty(AnatomicalAxesNotDef)
                checkArea = nan(size(PortionToCheck,1),size(AnatomicalAxesNotDef,1));
            else
                checkArea = nan(size(PortionToCheck,1),1);
            end
            for k = 1:size(PortionToCheck,1)
                for turns = 1:size(AnatomicalAxesNotDef,1)
                    checkArea(k,turns) = isempty(intersect(PortionToCheck(k,1):PortionToCheck(k,2),AnatomicalAxesNotDef(turns,1):AnatomicalAxesNotDef(turns,2)));
                end
            end
            if ~isempty(checkArea)
                if size(checkArea,2)>1
                    checkArea(:,end+1) = zeros(size(checkArea,1),1);
                    for k = 1:size(size(checkArea,1))
                        if ~isempty(find(checkArea(k,:), 1))
                            checkArea(k,end)=1;
                        else
                            checkArea(k,end)=0;
                        end
                    end
                    checkArea = checkArea(:,end);
                end
            end
            
            % area(s) of interest are those found outside the segments with gaps
            pos_N = find(checkArea);
            h = 20;
            for p = 1:length(pos_N)
                %check GEs in this area(s)
                IC(:,2) = ones(size(IC,1),1);
                for k = 1:size(IC,1)
                    if ~isempty(intersect(IC(k,1),PortionToCheck(p,1)-h:PortionToCheck(p,2)+h))
                        GE_new = GE_gh.HS;
                        [~,pos] = min(abs(IC(k,1)-GE_new));
                        if abs(IC(k,1)-GE_new(pos))<twindow+h
                            IC(k,1) = GE_new(pos);
                        else
                            IC(k,2) = 0;
                        end
                    end
                end
                IC = IC(IC(:,2)==1);
                
                FC(:,2) = ones(size(FC,1),1);
                for k = 1:length(FC)
                    if ~isempty(intersect(FC(k),PortionToCheck(p,1)-h:PortionToCheck(p,2)+h))
                        GE_new = GE_gh.TO;
                        [~,pos] = min(abs(FC(k)-GE_new));
                        if abs(FC(k,1)-GE_new(pos))<twindow+h
                            FC(k,1) = GE_new(pos);
                        else
                            FC(k,2) = 0;
                        end
                    end
                end
                FC = FC(FC(:,2)==1);
            end
        end % check portion of the signal with spikes in foot velocity --> possible mrk swap!
        %%
        
        %% Remove events found where Anatomical axes cannot be assessed
        h = 50;
        if ~isempty(AnatomicalAxesNotDef)
            for tr = 1:size(AnatomicalAxesNotDef,1)
                IC(:,2) = ones(size(IC,1),1);
                for k = 1:size(IC,1)
                    if ~isempty(intersect(IC(k,1),AnatomicalAxesNotDef(tr,1)-h:AnatomicalAxesNotDef(tr,2)+h))
                        GE_new = GE_gh.HS;
                        [~,pos] = min(abs(IC(k,1)-GE_new));
                        if abs(IC(k,1)-GE_new(pos))<twindow+h
                            IC(k,1) = GE_new(pos);
                        else
                            IC(k,2) = 0;
                        end
                    end
                end
                IC = IC(IC(:,2)==1);
                IC = unique(IC);
                
                FC(:,2) = ones(size(FC,1),1);
                for k = 1:size(FC,1)
                    if ~isempty(intersect(FC(k),AnatomicalAxesNotDef(tr,1)-h:AnatomicalAxesNotDef(tr,2)+h))
                        GE_new = GE_gh.TO;
                        [~,pos] = min(abs(FC(k)-GE_new));
                        if abs(FC(k,1)-GE_new(pos))<twindow+h
                            FC(k,1) = GE_new(pos);
                        else
                            FC(k,2) = 0;
                        end
                    end
                end
                FC = FC(FC(:,2)==1);
                FC = unique(FC);
            end
        end
        
        %% Merge other ICs
        IC_temp = IC;
        for k = 1:size(IC,1)
            posNow = find(GE_gh.HS<IC(k));
            if ~isempty(posNow)
                candidateICs = GE_gh.HS(posNow);
                if k>1
                    refPos = find((IC(k)-candidateICs)>50 & abs(IC(k-1)-candidateICs)>50);
                    IC_temp = [IC_temp; GE_gh.HS(posNow(refPos))]; IC_temp = unique(IC_temp);
                else
                    refPos = find((IC(k)-candidateICs)>50);
                    IC_temp = [IC_temp; GE_gh.HS(posNow(refPos))]; IC_temp = unique(IC_temp);
                end
            end
            if k == size(IC,1)
                posNow = find(GE_gh.HS>IC(k));
                if ~isempty(posNow)
                    candidateICs = GE_gh.HS(posNow);
                    refPos = find(abs(IC(k)-candidateICs)>50);
                    IC_temp = [IC_temp; GE_gh.HS(posNow(refPos))]; IC_temp = unique(IC_temp);
                end
            end
        end
        IC = IC_temp; clear ICtemp;
        
        %% CHECK HS/TO
        % for each FC there should be an IC before
        if ~isempty(IC)
            for k = 1:size(FC,1)
                if k ==1
                    IC_now = intersect(IC(:,1),1:FC(k,1));
                else
                    IC_now = intersect(IC(:,1),FC(k-1,1):FC(k,1));
                end
                if length(IC_now)>1 % too many ICs found
                    for m = 1:length(IC_now)
                        if isempty(find(abs(IC_now(m) - IC)<3*twindow, 1))
                            IC = [IC; IC_now];
                        end
                    end
                elseif isempty(IC_now) % FC NOT found
                    GE_new = GE_gh.HS;
                    if ~isempty(GE_new)
                        if k ==1
                            IC_now = intersect(GE_new(:,1),1:FC(k,1));
                        else
                            IC_now = intersect(GE_new(:,1),FC(k-1,1):FC(k,1));
                        end
                        if length(IC_now)==1
                            if isempty(find(abs(IC_now - IC)<3*twindow, 1))
                                IC = [IC; IC_now];
                            end
                        end
                    end
                end
            end
        end
        IC = sort(IC);
        
        % for each IC there should be a FC afterwards
        for k = 1:size(IC,1)-1
            FC_now = intersect(FC(:,1),IC(k,1):IC(k+1,1));
            if length(FC_now)>1 % too many FCs found
                for m = 1:length(FC_now)
                    if isempty(find(abs(FC_now(m) - FC)<3*twindow, 1))
                        FC = [FC; FC_now];
                    end
                end
            elseif isempty(FC_now) % FC NOT found
                GE_new = GE_gh.TO;
                FC_now = intersect(GE_new(:,1),IC(k,1):IC(k+1,1));
                if length(FC_now)==1
                    if isempty(find(abs(FC_now - FC)<3*twindow, 1))
                        FC = [FC; FC_now];
                    end
                end
            end
        end
        FC = sort(FC);
        
    else
        
    end
    
    %% Figure
    if fig
        time = 1/fs:1/fs:N/fs;
        figure('Name',strcat(side,'_GE detection'),'NumberTitle','off')
        subplot(311)
        hold on
        plot(time,TrajH, 'b')
        plot(time, TrajT, 'r')
        minV = min([min(TrajH),min(TrajT)]);
        maxV = max([max(TrajH),max(TrajT)]);
        for i = 1:size(AnatomicalAxesNotDef,1)
            X2=([AnatomicalAxesNotDef(i,1)/fs, AnatomicalAxesNotDef(i,2)/fs]);
            Y2=([minV,minV]);
            Y3=([maxV,maxV]);
            I = patch([X2 fliplr(X2)],[Y2 fliplr(Y3)], 'b', 'EdgeColor','none');
            alpha(0.4);
        end
        if ~isempty(IC)
            scatter(IC/fs, TrajH(IC,1),'filled','o','MarkerFaceColor','cyan','MarkerEdgeColor','blue')
            text((IC(1)/fs),TrajH(IC(1,1)),'IC')
        end
        if  ~isempty(FC)
            scatter(FC/fs, TrajT(FC,1), 'filled','o','MarkerFaceColor','green')
            text((FC(1)/fs),TrajT(FC(1,1)),'FC')
        end
        ylabel('AP marker displacement [m]')
        
        subplot(312)
        hold on
        plot(time(1:end-1), Hvelocity,'b')
        plot(time(1:end-1), Tvelocity,'r')
        minV = min([min(Data.(headerTraj).(strcat(side,FootMrk{1}))(:,3)),...
            min(Data.(headerTraj).(strcat(side,FootMrk{2}))(:,3))]);
        maxV = max([max(Data.(headerTraj).(strcat(side,FootMrk{1}))(:,3)),...
            max(Data.(headerTraj).(strcat(side,FootMrk{2}))(:,3))]);
        for i = 1:size(AnatomicalAxesNotDef,1)
            X2=([AnatomicalAxesNotDef(i,1)/fs, AnatomicalAxesNotDef(i,2)/fs]);
            Y2=([minV,minV]);
            Y3=([maxV,maxV]);
            I = patch([X2 fliplr(X2)],[Y2 fliplr(Y3)], 'b', 'EdgeColor','none');
            alpha(0.1);
        end
        if ~isempty(FC)
            scatter((FC/fs),Tvelocity(FC),'filled','o','MarkerFaceColor','green')
            text((FC(1)/fs),Tvelocity(FC(1)),'FC')
        end
        if ~isempty(IC)
            scatter((IC/fs),Hvelocity(IC),'filled','o',...
                'MarkerFaceColor','cyan','MarkerEdgeColor','blue')
            text((IC(1)/fs),Hvelocity(IC(1)),'IC')
        end
        ylabel('3D marker velocity [m/s]')
        
        subplot(313)
        hold on
        plot(time, Data.(headerTraj).(strcat(side,FootMrk{1}))(:,3),'b')
        plot(time, Data.(headerTraj).(strcat(side,FootMrk{2}))(:,3),'r')
        minV = min([min(Data.(headerTraj).(strcat(side,FootMrk{1}))(:,3)),...
            min(Data.(headerTraj).(strcat(side,FootMrk{2}))(:,3))]);
        maxV = max([max(Data.(headerTraj).(strcat(side,FootMrk{1}))(:,3)),...
            max(Data.(headerTraj).(strcat(side,FootMrk{2}))(:,3))]);
        for i = 1:size(AnatomicalAxesNotDef,1)
            X2=([AnatomicalAxesNotDef(i,1)/fs, AnatomicalAxesNotDef(i,2)/fs]);
            Y2=([minV,minV]);
            Y3=([maxV,maxV]);
            I = patch([X2 fliplr(X2)],[Y2 fliplr(Y3)], 'b', 'EdgeColor','none');
            alpha(0.1);
        end
        if ~isempty(FC)
            scatter((FC/fs),Data.(headerTraj).(strcat(side,FootMrk{2}))(FC,3),'filled','o','MarkerFaceColor','green')
            text((FC(1)/fs),Data.(headerTraj).(strcat(side,FootMrk{2}))(FC(1),3),'FC')
        end
        if ~isempty(IC)
            scatter((IC/fs),Data.(headerTraj).(strcat(side,FootMrk{1}))(IC,3),'filled','o',...
                'MarkerFaceColor','cyan','MarkerEdgeColor','blue')
            text((IC(1)/fs),Data.(headerTraj).(strcat(side,FootMrk{1}))(IC(1),3),'IC')
        end
        
        ylabel('Vertical marker trajectories [m]')
        xlabel('Time[s]')
        l = legend(FootMrk{1},FootMrk{2},'Location','southoutside');
        l.NumColumns = 2;
    end
    
else
    warning('Check data unit! \n')
    return
end
end